import * as CSS from 'csstype';
import { Options, Meta } from './typings';
export { Options, Meta };
/**
 * @example p2,mx4,left10,gap4...
 * @example p-20,opacity-80
 */
export declare const digitReg: RegExp;
export declare class Atom {
    private readonly options;
    private readonly autoSetId;
    /**
     *
     * @param options
     * @param autoSetId  setId when new Atom(), default is true
     */
    constructor(options: Options, autoSetId?: boolean);
    /**
     * string or number prop
     * @readonly
     * @memberof Atom
     */
    get isValueProp(): boolean;
    /**
     * get style keys string
     * @readonlyReactElement
     * @memberof Atom
     * @example
     * { color: "#999"} -> color
     * { paddingTop: 10, paddingBottom: 10} -> paddingTop-paddingTop
     */
    get styleKeys(): string;
    /**
     * get style key hash
     * @readonly
     * @memberof Atom
     * @example
     * { color: "#999"} -> color + hashed meta
     * { paddingTop: 10, paddingBottom: 10} -> paddingTop-paddingTop  + hashed meta
     */
    get styleKeysHash(): string;
    get isFalsyPropValue(): boolean;
    get isTruePropValue(): boolean;
    /**
     * original propKey, 原始的propkey
     * @example
     * <Box red200></Box> propKey is red200
     * <Box red200--hover></Box> propKey is red200--hover
     */
    readonly propKey: 'css' | 'debug' | ({} & string);
    /**
     * original propValue
     * @example
     * <Box red200></Box>  -> true
     * <Box p={10}></Box>  -> 10
     * <Box color="red"></Box>  -> red
     */
    readonly propValue: any;
    props: any;
    /**
     * all propKeys for this atom
     */
    propKeys: string[];
    /**
     * get the primitive atomic key, exclude value or postfix
     * @example
     * m-4 -> m
     * m={4} -> m
     * m4 -> m
     * m4--hover -> m
     * m4--sm -> m
     * m4--dark--sm-i -> m
     */
    key: string;
    /**
     * atom value, get it from PropValue or propKey
     * @example
     * <Box toCenter></Box> -> true
     * <Box m={4}></Box> -> 4
     * <Box m-4></Box> -> 4
     * <Box m4></Box> -> 16
     */
    value: any;
    meta: Meta;
    /**
     * unique id for cache
     */
    id: string;
    style: CSS.Properties<number | string>;
    type: 'color' | 'backgroundColor' | 'borderColor' | 'padding' | 'margin' | ({} & string);
    /**
     * if handled, this atom is ready to push to parser.atoms
     */
    handled: boolean;
    /**
     * already inserted to stylesheet
     */
    inserted: boolean;
    /** if not valid, do not generate style */
    isValid: boolean;
    setId: () => string;
    /**
     *
     * @param config fower config
     * @returns
     */
    preprocessAtom(): this;
    postfixPreprocessor(): this;
    digitPreprocessor(): this;
    private getInitialHandled;
    private getIsValid;
}
