import deepmerge from 'deepmerge';
import { isBrowser, isEmptyObj, upFirst, isPercentNumber, isNumber, jsKeyToCssKey, objectToClassName, argsToProps, hash as hash$1 } from '@fower/utils';
import { parse } from '@fower/css-object-processor';
import { formatColor } from '@fower/color-helper';
import hash from 'string-hash';

function _defineProperties(target, props) {
  for (var i = 0; i < props.length; i++) {
    var descriptor = props[i];
    descriptor.enumerable = descriptor.enumerable || false;
    descriptor.configurable = true;
    if ("value" in descriptor) descriptor.writable = true;
    Object.defineProperty(target, _toPropertyKey(descriptor.key), descriptor);
  }
}
function _createClass(Constructor, protoProps, staticProps) {
  if (protoProps) _defineProperties(Constructor.prototype, protoProps);
  if (staticProps) _defineProperties(Constructor, staticProps);
  Object.defineProperty(Constructor, "prototype", {
    writable: false
  });
  return Constructor;
}
function _extends() {
  _extends = Object.assign ? Object.assign.bind() : function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];
      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }
    return target;
  };
  return _extends.apply(this, arguments);
}
function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;
  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }
  return target;
}
function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}
function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;
  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];
  return arr2;
}
function _createForOfIteratorHelperLoose(o, allowArrayLike) {
  var it = typeof Symbol !== "undefined" && o[Symbol.iterator] || o["@@iterator"];
  if (it) return (it = it.call(o)).next.bind(it);
  if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") {
    if (it) o = it;
    var i = 0;
    return function () {
      if (i >= o.length) return {
        done: true
      };
      return {
        done: false,
        value: o[i++]
      };
    };
  }
  throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}
function _toPrimitive(input, hint) {
  if (typeof input !== "object" || input === null) return input;
  var prim = input[Symbol.toPrimitive];
  if (prim !== undefined) {
    var res = prim.call(input, hint || "default");
    if (typeof res !== "object") return res;
    throw new TypeError("@@toPrimitive must return a primitive value.");
  }
  return (hint === "string" ? String : Number)(input);
}
function _toPropertyKey(arg) {
  var key = _toPrimitive(arg, "string");
  return typeof key === "symbol" ? key : String(key);
}

var modeCacheKey = 'fower-mode';
var Store = /*#__PURE__*/function () {
  function Store() {
    var _this = this;
    this.config = {
      unit: 'px',
      objectPropKeys: ['css'],
      inline: false,
      important: false,
      mode: {
        currentMode: 'light',
        autoDarkMode: {
          enabled: false,
          mappings: {
            white: 'black',
            black: 'white',
            '50': '900',
            '100': '800',
            '200': '700',
            '300': '600',
            '400': '500',
            '500': '400',
            '600': '300',
            '700': '200',
            '800': '100',
            '900': '50'
          }
        },
        modeList: ['light', 'dark'],
        classPrefix: ''
      },
      prefix: '',
      pseudos: ['active', 'checked', 'disabled', 'enabled', 'default', 'empty', 'focus', 'focus-within', 'invalid', 'hover', 'link', 'visited', 'first-child', 'last-child', 'after', 'before', 'placeholder', 'selection'],
      theme: {
        breakpoints: {},
        colors: {},
        spacings: {}
      },
      plugins: []
    };
    this.atomCache = new Map();
    this.getAtomIds = function () {
      return Array.from(_this.atomCache.keys());
    };
    // composed atomic props
    this.compositions = new Map();
    this.getConfig = function () {
      return _this.config;
    };
    // user config
    this.setConfig = function (config, strategy) {
      if (strategy === void 0) {
        strategy = 'deepmerge';
      }
      if (strategy === 'replace') {
        _this.config = config;
      } else if (strategy === 'merge') {
        _this.config = _extends({}, _this.config, config);
      } else {
        _this.config = deepmerge(_this.config, config);
        // TODO: need improve, this a special key
        if (config.objectPropKeys) {
          _this.config.objectPropKeys = config.objectPropKeys;
        }
      }
    };
    this.getTheme = function () {
      return _this.config.theme;
    };
    this.setTheme = function (partialThemeConfig) {
      _this.config.theme = deepmerge(_this.config.theme || {}, partialThemeConfig);
    };
    this.getMode = function () {
      var _this$config$mode;
      return ((_this$config$mode = _this.config.mode) == null ? void 0 : _this$config$mode.currentMode) || '';
    };
    this.setMode = function (mode) {
      if (!isBrowser) return;
      var currentMode = _this.config.mode.currentMode;
      if (currentMode) {
        document.documentElement.classList.remove(currentMode);
      }
      if (mode) {
        document.documentElement.classList.add(mode);
      }
      localStorage.setItem(modeCacheKey, mode);
      _this.setConfig({
        mode: {
          currentMode: mode
        }
      });
    };
    this.use = function () {
      var _this$config$plugins;
      (_this$config$plugins = _this.config.plugins).push.apply(_this$config$plugins, arguments);
    };
    this.addAtom = function (matcher, handleAtomOrStyleObject) {
      var plugin = {
        isMatch: function isMatch(key) {
          if (typeof matcher === 'string') return key === matcher;
          if (matcher instanceof RegExp) return matcher.test(key);
          return false;
        },
        handleAtom: typeof handleAtomOrStyleObject === 'function' ? handleAtomOrStyleObject : function (atom) {
          atom.style = handleAtomOrStyleObject;
          return atom;
        }
      };
      _this.use(plugin);
      return plugin;
    };
    this.composeAtom = function (atomName, cssObject) {
      _this.compositions.set(atomName, cssObject);
    };
  }
  _createClass(Store, [{
    key: "theme",
    get: function get() {
      return this.config.theme;
    }
  }]);
  return Store;
}();
var store = /*#__PURE__*/new Store();

var _excluded = ["pseudoPrefix", "childSelector", "important"],
  _excluded2 = ["colorPostfix"];
var invalidProps = ['excludedProps'];
/**
 * @example p2,mx4,left10,gap4...
 * @example p-20,opacity-80
 */
var digitReg = /^(m[xytrbl]?-?|p[xytrbl]?|gap[xy]?|space[xy]?|top-?|right-?|bottom-?|left-?|[wh]|square|circle|min[hw]|max[hw]|opacity|delay|duration|translate[xyz]|scale[xy]?|rotate[xy]?|skew[xy]?|text|zIndex-?|leading|stroke|fontWeight|outlineOffset|order|flex(Grow|Shrink|Basis)?|(row|column)?Gap|gridTemplateColumns|border(Top|Right|Bottom|Left)?|rounded(Top(Left|Right)?|Right|Bottom(Left|Right)?|Left)?)(-?\d+[a-z]*?|-auto)$/i;
var Atom = /*#__PURE__*/function () {
  /**
   *
   * @param options
   * @param autoSetId  setId when new Atom(), default is true
   */
  function Atom(options, autoSetId) {
    var _this = this;
    if (autoSetId === void 0) {
      autoSetId = true;
    }
    this.options = options;
    this.autoSetId = autoSetId;
    this.props = {};
    /**
     * all propKeys for this atom
     */
    this.propKeys = [];
    /**
     * atom value, get it from PropValue or propKey
     * @example
     * <Box toCenter></Box> -> true
     * <Box m={4}></Box> -> 4
     * <Box m-4></Box> -> 4
     * <Box m4></Box> -> 16
     */
    this.value = '';
    /**
     * unique id for cache
     */
    this.id = '';
    this.type = '';
    this.setId = function () {
      var meta = _this.meta,
        key = _this.key,
        value = _this.value;
      var important = meta.important,
        rest = _objectWithoutPropertiesLoose(meta, _excluded);
      var values = Object.values(rest).sort();
      if (important) values.push('i');
      var id = '';
      // is global style
      if (meta.global) {
        id = hash(JSON.stringify(value)).toString();
        _this.id = id;
        return id;
      }
      if (typeof value === 'boolean' && value === true) {
        id = key;
      } else if (Array.isArray(value)) {
        var valueStr = value.join('-');
        id = key + "-" + valueStr;
      } else if (typeof value === 'function') {
        id = key + "-" + value(_this.props);
      } else {
        id = key + "-" + String(value);
      }
      if (values.length) id = id + '--' + values.join('--');
      // handle special character
      id = id.replace(/[#()"']/g, '').replace('.', '_').replace(/\%/g, 'p').replace(/\s+/g, '-');
      if (_this.isFalsyPropValue) id = id + '--false';
      if (meta.childSelector) {
        id = meta.childSelector + '-' + id;
      }
      var isValid = /^[a-zA-Z0-9-_]+$/.test(id);
      id = isValid ? id : "css-" + hash(id);
      _this.id = id;
      return id;
    };
    this.propKey = options.propKey;
    this.propValue = options.propValue;
    this.propKeys = [this.propKey];
    this.key = options.key || this.propKey;
    this.value = options.value || this.propValue;
    this.style = options.style || {};
    this.props = options.props || {};
    this.handled = this.getInitialHandled();
    this.isValid = this.getIsValid();
    this.inserted = false;
    // shallow clone it
    this.meta = _extends({}, options.meta) || {};
    this.type = options.type || '';
    this.preprocessAtom();
  }
  /**
   * string or number prop
   * @readonly
   * @memberof Atom
   */
  var _proto = Atom.prototype;
  /**
   *
   * @param config fower config
   * @returns
   */
  _proto.preprocessAtom = function preprocessAtom() {
    var newAtom = this.postfixPreprocessor();
    if (this.autoSetId) this.setId();
    return newAtom;
  };
  _proto.postfixPreprocessor = function postfixPreprocessor() {
    var connector = '--';
    var specialPseudos = ['after', 'before', 'placeholder', 'selection'];
    var _store$config = store.config,
      _store$config$pseudos = _store$config.pseudos,
      pseudos = _store$config$pseudos === void 0 ? [] : _store$config$pseudos,
      theme = _store$config.theme,
      mode = _store$config.mode;
    var modeList = mode.modeList;
    var breakpoints = theme.breakpoints;
    var propKey = this.propKey,
      propValue = this.propValue;
    var breakpointKeys = Object.keys(breakpoints);
    var modeKeys = modeList || [];
    var pseudoKeys = pseudos;
    var regResponsiveStr = connector + "(" + breakpointKeys.join('|') + ")";
    var regModeStr = connector + "(" + modeKeys.join('|') + ")";
    var regPseudoStr = connector + "(" + pseudoKeys.join('|') + ")";
    var regMode = new RegExp(regModeStr);
    var regPseudo = new RegExp(regPseudoStr);
    var regResponsive = new RegExp(regResponsiveStr);
    var regImportant = /--i/i;
    var regColorPostfix = /--[told](\d{1,2}|100)($|--)/i;
    var regParentSelector = /--\$\w+/i;
    /** handle value like: red500--T40, #666--O30 */
    if (regColorPostfix.test(propValue)) {
      var _propValue$split = propValue.split('--'),
        colorName = _propValue$split[0],
        postfix = _propValue$split[1];
      this.value = colorName;
      this.meta.colorPostfix = postfix;
    }
    var isMode = regMode.test(propKey);
    var isPseudo = regPseudo.test(propKey);
    var isResponsive = regResponsive.test(propKey);
    var isImportant = regImportant.test(propKey);
    var isColorPostfix = regColorPostfix.test(propKey);
    var isParentSelector = regParentSelector.test(propKey);
    var hasPostfix = isMode || isPseudo || isResponsive || isImportant || isColorPostfix;
    if (!hasPostfix) return this.digitPreprocessor();
    var result = propKey.split(connector);
    this.key = result[0]; // key that already removed postfix
    if (isMode) {
      this.meta.mode = result.find(function (i) {
        return modeKeys.includes(i);
      });
    }
    if (isPseudo) {
      var pseudo = result.find(function (i) {
        return pseudoKeys.includes(i);
      });
      this.meta.pseudoPrefix = specialPseudos.includes(pseudo) ? '::' : ':';
      this.meta.pseudo = pseudo;
    }
    if (isResponsive) {
      var breakpointType = result.find(function (i) {
        return breakpointKeys.includes(i);
      });
      this.meta.breakpoint = breakpointType;
    }
    if (isImportant) {
      this.meta.important = !!result.find(function (i) {
        return i === 'i';
      });
    }
    if (isColorPostfix) {
      this.meta.colorPostfix = result.find(function (i) {
        return regColorPostfix.test("--" + i);
      });
    }
    if (isParentSelector) {
      var _result$find;
      this.meta.parentClass = (_result$find = result.find(function (i) {
        return i.startsWith('$');
      })) == null ? void 0 : _result$find.replace(/^\$/, '');
    }
    // check is theme space key, if yes, preprocess it
    this.digitPreprocessor();
    return this;
  };
  _proto.digitPreprocessor = function digitPreprocessor() {
    if (!digitReg.test(this.key)) return this;
    var spacings = store.config.theme.spacings;
    // is theme space key
    var isSpace = /^([a-z]+)(\d+)$/i.test(this.key);
    /**
     *  match props link: m4,mx2,mt9, gap4...
     *  to m4 -> [ key: m, value: 4 ]
     *  to m-20 -> [ key: m, value: 20 ]
     *  to m-20px -> [ key: m, value: '20px' ]
     */
    var keyStr = this.key.toString();
    var result = keyStr.match(/^([a-z]+)(\d+)$/i) || keyStr.match(/^([a-z]*)-(-?\d+[a-z]*?)$/i) || keyStr.match(/^([a-z]+)-(auto)$/i);
    if (!result) return this;
    var newKey = result[1],
      newPropValue = result[2];
    this.key = newKey;
    this.value = isSpace ? spacings["" + newPropValue.toLowerCase()] : newPropValue;
    return this;
  };
  _proto.getInitialHandled = function getInitialHandled() {
    if (this.options.handled) return this.options.handled;
    if (this.isFalsyPropValue) return true;
    if (invalidProps.includes(this.propKey)) return true;
    return false;
  };
  _proto.getIsValid = function getIsValid() {
    if (this.isFalsyPropValue) return false;
    if (invalidProps.includes(this.propKey)) return false;
    return true;
  };
  _createClass(Atom, [{
    key: "isValueProp",
    get: function get() {
      return typeof this.propValue === 'string' || typeof this.propValue == 'number';
    }
    /**
     * get style keys string
     * @readonlyReactElement
     * @memberof Atom
     * @example
     * { color: "#999"} -> color
     * { paddingTop: 10, paddingBottom: 10} -> paddingTop-paddingTop
     */
  }, {
    key: "styleKeys",
    get: function get() {
      return Object.keys(this.style || {}).join('-');
    }
    /**
     * get style key hash
     * @readonly
     * @memberof Atom
     * @example
     * { color: "#999"} -> color + hashed meta
     * { paddingTop: 10, paddingBottom: 10} -> paddingTop-paddingTop  + hashed meta
     */
  }, {
    key: "styleKeysHash",
    get: function get() {
      var _this$meta = this.meta,
        rest = _objectWithoutPropertiesLoose(_this$meta, _excluded2); // omit colorPostfix
      return Object.keys(this.style || {}).join('-') + JSON.stringify(rest);
    }
  }, {
    key: "isFalsyPropValue",
    get: function get() {
      if (this.propValue === undefined || this.propValue === null) return true;
      return typeof this.propValue === 'boolean' && !this.propValue;
    }
  }, {
    key: "isTruePropValue",
    get: function get() {
      return typeof this.propValue === 'boolean' && this.propValue;
    }
  }]);
  return Atom;
}();

var StyleSheet = /*#__PURE__*/function () {
  function StyleSheet() {
    this.$style = null;
    this.cssString = '';
    this.ssrAtomIds = null;
    if (isBrowser) {
      var _globalThis$document;
      this.$style = (_globalThis$document = globalThis.document) == null || _globalThis$document.querySelector == null ? void 0 : _globalThis$document.querySelector('[data-fower]');
    }
  }
  var _proto = StyleSheet.prototype;
  _proto.createStyleElement = function createStyleElement() {
    var $style = document.createElement('style');
    $style.dataset.fower = '';
    document.head.append($style);
    return $style;
  };
  _proto.getSsrAtomIds = function getSsrAtomIds() {
    var _this$$style;
    if (!this.$style) return [];
    if (this.ssrAtomIds) return this.ssrAtomIds;
    return ((_this$$style = this.$style) == null || (_this$$style = _this$$style.dataset.fower) == null ? void 0 : _this$$style.split(',')) || [];
  };
  _proto.insertStyleToHtml = function insertStyleToHtml(rules) {
    if (!isBrowser) return;
    if (!this.$style) {
      this.$style = this.createStyleElement();
    }
    // const str = rules.join('\n')
    // this.$style.innerHTML += str
    // return
    for (var _iterator = _createForOfIteratorHelperLoose(rules), _step; !(_step = _iterator()).done;) {
      var rule = _step.value;
      try {
        this.$style.sheet.insertRule(rule);
      } catch (error) {
        console.warn(error);
      }
    }
  };
  _proto.insertStyles = function insertStyles(rules) {
    if (rules === void 0) {
      rules = [];
    }
    if (!rules.length) return;
    if (isBrowser) return this.insertStyleToHtml(rules);
    this.cssString = this.cssString + ' ' + rules.join(' ');
  };
  _proto.getStyle = function getStyle() {
    return this.cssString;
  };
  return StyleSheet;
}();
var styleSheet = /*#__PURE__*/new StyleSheet();

//  from https://github.com/modulz/stitches/blob/canary/packages/core/src/unitOnlyProps.js
var reg = /(-columns|(^|[^e]-)padding|[^t]-spacing|l-align|rows|(^|(^border|[dkmnptx]|le|ne)-)width|^border|tom|[ek]-start|(o|[^e]-du)ration|us|(^|[^tv]-)left|(^|-)top|tance|rgin|e-offset|(er|g|n|t)-block|(^|[^tv]-)right|basis|[gnt]-inline|gap|(^|[^e]-)height|ness|(^|[^p]-)inset|[ek]-end|elay|tline|ve|dent|-rotate|n-rule|(c|ck|d|ne|t)-size)$/;
/**
 * check css need unit
 * @param cssKey css key, eg: border-width
 * @returns
 */
function isUnitProp(cssKey) {
  return reg.test(cssKey);
}

var colorKeys = ['color', 'backgroundColor', 'borderColor'];
//  high-frequency used props in react
var reactProps = ['children', 'onClick', 'onChange', 'onBlur', 'id', 'className', 'style', 'title', 'tabIndex', 'placeholder', 'href', 'src', 'contentEditable', 'defaultChecked', 'defaultValue'];
/**
 * An Abstract tool to handle atomic props
 */
var Parser = /*#__PURE__*/function () {
  function Parser(props) {
    var _store$config$mode;
    if (props === void 0) {
      props = {};
    }
    this.props = props;
    /**
     * atom parsed from props
     */
    this.atoms = [];
    /**
     * For extensibility
     */
    this.data = {};
    this.getClassNameById = function (id) {
      return store.config.prefix + id;
    };
    this.traverseProps(props);
    if ((_store$config$mode = store.config.mode) != null && (_store$config$mode = _store$config$mode.autoDarkMode) != null && _store$config$mode.enabled) {
      this.createAutoDarkModeAtom();
    }
  }
  var _proto = Parser.prototype;
  /**
   * traverse Props to init atoms
   */
  _proto.traverseProps = function traverseProps(props) {
    if (isEmptyObj(props)) return;
    var _props$excludedProps = props.excludedProps,
      excludedProps = _props$excludedProps === void 0 ? [] : _props$excludedProps;
    var entries = Object.entries(props);
    for (var _iterator = _createForOfIteratorHelperLoose(this.plugins), _step; !(_step = _iterator()).done;) {
      var plugin = _step.value;
      if (plugin.init) {
        plugin.init(props);
      }
    }
    // traverse Props
    for (var _i = 0, _entries = entries; _i < _entries.length; _i++) {
      var _this$config$objectPr;
      var _entries$_i = _entries[_i],
        propKey = _entries$_i[0],
        propValue = _entries$_i[1];
      // the prop should be excluded by user setting
      if (excludedProps.includes(propKey)) continue;
      if (reactProps.includes(propKey)) continue;
      if (!this.isValidProp(propKey, propValue)) continue;
      // parse object prop
      if ((_this$config$objectPr = this.config.objectPropKeys) != null && _this$config$objectPr.includes(propKey)) {
        this.parseObjectProp(propValue, {});
        continue;
      }
      if (Array.isArray(propValue)) {
        // TODO: hack for bgGradientX, bgGradientY
        if (/^bgGradient[XY].*/i.test(propKey) && !Array.isArray(propValue[0])) ; else {
          this.parseResponsiveValue(propKey, propValue);
          continue;
        }
      }
      var composition = store.compositions.get(propKey);
      if (composition) {
        this.parseObjectProp(composition, {});
        var _atom = new Atom({
          propKey: propKey,
          propValue: propValue,
          props: this.props
        });
        _atom.handled = true;
        _atom.style = {};
        this.addAtom(_atom);
        continue;
      }
      // common props
      var atom = new Atom({
        propKey: propKey,
        propValue: propValue,
        props: this.props
      }, false);
      try {
        this.mutateAtom(atom);
        if (atom.handled) this.addAtom(atom);
      } catch (error) {
        continue;
      }
    }
    for (var _iterator2 = _createForOfIteratorHelperLoose(this.plugins), _step2; !(_step2 = _iterator2()).done;) {
      var _plugin = _step2.value;
      if (_plugin.afterAtomStyleCreate) {
        _plugin.afterAtomStyleCreate(this);
      }
    }
  };
  _proto.getAutoDarkModeAtom = function getAutoDarkModeAtom(atom) {
    var _options$style;
    if (atom.meta.mode) return null;
    if (!atom.style) return null;
    var entries = Object.entries(atom.style);
    if (!(entries != null && entries.length)) return null;
    var options = {};
    var darkColor = '';
    var config = store.config;
    var mappings = config.mode.autoDarkMode.mappings;
    var colors = config.theme.colors;
    var _entries$ = entries[0],
      colorKey = _entries$[0],
      colorName = _entries$[1];
    // check is valid color key
    if (!colorKeys.includes(colorKey)) return null;
    // only color in theme is valid
    if (!colors[colorName]) return null;
    // if scale is existed, scale should be 100,200,300,400...
    var _ref = colorName.match(/^([a-z]+)(\d+)$/i) || [],
      scale = _ref[2];
    // is disable, eg: { red200: false}
    if (mappings[colorName] === false) return null;
    if (mappings[atom.propKey]) {
      darkColor = mappings[atom.propKey];
    } else if (mappings[colorName]) {
      darkColor = mappings[colorName];
    } else {
      darkColor = colorName.replace(scale, mappings[scale]);
    }
    var isColor = colorKey === 'color';
    var isBgColor = colorKey === 'backgroundColor';
    var isBorderColor = colorKey === 'borderColor';
    if (colors[darkColor]) {
      if (isColor) {
        options.propKey = darkColor;
        options.type = 'color';
      } else if (isBgColor) {
        options.propKey = "bg" + upFirst(darkColor);
        options.type = 'backgroundColor';
      } else if (isBorderColor) {
        options.propKey = "border" + upFirst(darkColor);
        options.type = 'borderColor';
      }
      options.propValue = true;
    } else {
      if (isColor) {
        options.propKey = 'color';
        options.propValue = darkColor;
      } else if (isBgColor) {
        options.propKey = 'bg';
        options.propValue = darkColor;
      } else if (isBorderColor) {
        options.propKey = 'borderColor';
        options.propValue = darkColor;
      }
    }
    options.style = (_options$style = {}, _options$style[colorKey] = darkColor, _options$style);
    return new Atom(_extends({}, options, {
      props: this.props,
      meta: _extends({}, atom.meta, {
        mode: 'dark'
      })
    }));
  };
  _proto.createAutoDarkModeAtom = function createAutoDarkModeAtom() {
    var _this = this;
    var _loop = function _loop() {
        var _atom$meta;
        var atom = _step3.value;
        // no need auto dark mode
        if (!atom.isValid || ((_atom$meta = atom.meta) == null ? void 0 : _atom$meta.mode) === 'dark' || !colorKeys.includes(atom.type)) return 0; // continue
        // 已经存在 dark mode 的样式，不需要自动生成
        var find = _this.atoms.find(function (i) {
          return atom.type === i.type && i.meta.mode === 'dark' && !Reflect.has(i.meta, 'pseudo') && !Reflect.has(i.meta, 'childSelector');
        });
        if (find) return 0; // continue
        var darkAtom = _this.getAutoDarkModeAtom(atom);
        if (!darkAtom) return 0; // continue
        var cachedAtom = store.atomCache.get(darkAtom.id);
        if (cachedAtom) {
          _this.addAtom(cachedAtom);
        } else {
          _this.addAtom(darkAtom);
        }
      },
      _ret;
    for (var _iterator3 = _createForOfIteratorHelperLoose(this.atoms), _step3; !(_step3 = _iterator3()).done;) {
      _ret = _loop();
      if (_ret === 0) continue;
    }
  }
  /**
   * Get final css value
   * @param key css key, eg: font-size, padding-top
   * @param value css value
   * @returns
   */;
  _proto.formatCssValue = function formatCssValue(key, value) {
    // no need unit
    if (!isUnitProp(key)) return value;
    var numValue = value;
    // 80p -> 80%, 50p-> -50%
    if (isPercentNumber(String(value))) {
      return String(value).replace('p', '%');
    }
    if (!isNumber(value)) return value;
    numValue = Number(value);
    // if num is between 0 and 1, convert it to percent number.
    if (numValue < 1 && numValue > 0) {
      return numValue * 100 + '%';
    }
    var config = store.config;
    if (config.unit !== 'none') {
      if (config.transformUnit) {
        return config.transformUnit(numValue);
      } else {
        return value + store.config.unit;
      }
    }
    return numValue;
  }
  /**
   * convert style object to string
   * @param style
   * @param meta
   * @example
   * { width: 10 } -> "width: 10px;"
   * { paddingTop: 10, paddingBottom: 10 } -> "padding-top: 10px;padding-bottom: 10px;"
   * @returns
   */;
  _proto.styleToString = function styleToString(style, meta) {
    var _this2 = this;
    var important = meta.important,
      colorPostfix = meta.colorPostfix;
    return Object.entries(style).reduce(function (r, _ref2) {
      var key = _ref2[0],
        value = _ref2[1];
      // remove --sm,--md,--lg,--xl postfix
      key = key.replace(/\-\-.+/, '');
      var cssKey = jsKeyToCssKey(key);
      var postfix = important ? ' !important' : '';
      var colors = store.theme.colors;
      try {
        if (colorPostfix) {
          value = formatColor((colors == null ? void 0 : colors[value]) || value, colorPostfix);
        } else {
          value = _this2.formatCssValue(cssKey, (colors == null ? void 0 : colors[value]) || value);
        }
        return r + (cssKey + ": " + value + postfix + ";");
      } catch (error) {
        return '';
      }
    }, '');
  };
  _proto.addAtom = function addAtom(atom) {
    var _this$config$theme, _store$config$mode2;
    atom.props = this.props;
    if (!atom.id) atom.setId();
    // if not cached, let's cache it
    if (!store.atomCache.get(atom.id)) {
      store.atomCache.set(atom.id, atom);
    }
    var ssrAtomIds = styleSheet.getSsrAtomIds();
    var _ref3 = ((_this$config$theme = this.config.theme) == null ? void 0 : _this$config$theme.colors) || {},
      _ref3$modes = _ref3.modes,
      modes = _ref3$modes === void 0 ? {} : _ref3$modes;
    var entries = Object.entries(modes);
    /** for color mode */
    for (var _i2 = 0, _entries2 = entries; _i2 < _entries2.length; _i2++) {
      var _entries2$_i = _entries2[_i2],
        mode = _entries2$_i[0],
        colors = _entries2$_i[1];
      if (!atom.style) continue;
      var _entries3 = Object.entries(atom.style);
      if (!_entries3.length) continue;
      var _entries3$ = _entries3[0],
        styleKey = _entries3$[0],
        styleValue = _entries3$[1];
      var colorValue = colors[styleValue];
      if (colorValue) {
        var postfix = '--' + mode;
        // TODO: improve clone
        var modeAtom = JSON.parse(JSON.stringify(atom));
        modeAtom.key = atom.key + postfix;
        modeAtom.id = atom.id + postfix;
        modeAtom.meta = _extends({
          mode: mode
        }, atom.meta);
        modeAtom.style[styleKey] = colorValue;
        if (ssrAtomIds.includes(modeAtom.id)) modeAtom.inserted = true;
        this.atoms.push(modeAtom);
      }
    }
    if (ssrAtomIds.includes(atom.id)) atom.inserted = true;
    /**
     * handle color overwrite
     * 1. <Box red500 green500/>
     * 2. <Box red500 green500--hover/>
     */
    if ((_store$config$mode2 = store.config.mode) != null && (_store$config$mode2 = _store$config$mode2.autoDarkMode) != null && _store$config$mode2.enabled) {
      var _atom$meta2;
      if (atom.type === 'color' && ((_atom$meta2 = atom.meta) == null ? void 0 : _atom$meta2.mode) !== 'dark' && !Reflect.has(atom.meta, 'pseudo')) {
        var index = this.atoms.findIndex(function (i) {
          var _i$meta;
          return i.type === 'color' && ((_i$meta = i.meta) == null ? void 0 : _i$meta.mode) !== 'dark' && !Reflect.has(atom.meta, 'childSelector') && !Reflect.has(atom.meta, 'pseudo');
        });
        // 被覆盖的颜色设置为 isValid=false, 并且使用 copy 为新的 atom
        if (index > -1) {
          this.atoms[index] = _extends({}, this.atoms[index], {
            isValid: false
          });
        }
      }
    }
    this.atoms.push(atom);
  }
  /**
   * prop that can to handle, only primitive value type is valid
   * @param propKey
   * @param propValue
   * @returns
   */;
  _proto.isValidProp = function isValidProp(propKey, propValue) {
    var _this$config$objectPr2;
    var validTypes = ['string', 'boolean', 'number', 'undefined', 'function'];
    if ((_this$config$objectPr2 = this.config.objectPropKeys) != null && _this$config$objectPr2.includes(propKey)) return true;
    // for _hover,_sm,_dark...
    if (propKey.startsWith('_')) return true;
    if (Array.isArray(propValue)) return true;
    var type = typeof propValue;
    if (validTypes.includes(type)) return true;
    return false;
  }
  /**
   * to mutate atom attribute, and add atom to this.atoms
   * @param atom
   */;
  _proto.mutateAtom = function mutateAtom(atom) {
    for (var _iterator4 = _createForOfIteratorHelperLoose(this.plugins), _step4; !(_step4 = _iterator4()).done;) {
      var plugin = _step4.value;
      if (!(plugin.isMatch != null && plugin.isMatch(atom.key, this))) continue;
      // why set id here? because of function prop Value
      if (!atom.id) atom.setId();
      // handle function type props
      if (typeof atom.value === 'function') atom.value = atom.value(this.props);
      if (plugin.beforeHandleAtom) {
        atom = plugin.beforeHandleAtom(atom, this);
      }
    }
    // for nested style, like css props
    if (Object.values(atom.meta || {}).length || Object.values(atom.style || {}).length) {
      if (!atom.id) atom.setId();
    }
    var cachedAtom = store.atomCache.get(atom.id);
    if (cachedAtom) {
      if (!cachedAtom.propKeys.includes(atom.propKey)) {
        cachedAtom.propKeys.push(atom.propKey);
      }
      this.addAtom(cachedAtom);
      throw new Error('atom is cached, add to this.atoms directly, no need to mutate');
    }
    // if handled, push to this.atoms and skip it
    if (atom.handled) {
      this.addAtom(atom);
      throw new Error('atom is handled, add to this.atoms directly ,no need to mutate');
    }
    for (var _iterator5 = _createForOfIteratorHelperLoose(this.plugins), _step5; !(_step5 = _iterator5()).done;) {
      var _plugin2 = _step5.value;
      if (!(_plugin2.isMatch != null && _plugin2.isMatch(atom.key, this))) continue;
      if (_plugin2.handleAtom) {
        atom = _plugin2.handleAtom == null ? void 0 : _plugin2.handleAtom(atom, this);
      }
      atom.handled = true;
      break; // break from this plugin
    }
  };
  _proto.parseResponsiveValue = function parseResponsiveValue(propKey, propValue) {
    var breakpoints = this.config.theme.breakpoints;
    var keys = Object.keys(breakpoints);
    var obj = propValue.reduce(function (result, cur, i) {
      var _extends2;
      var prop = "" + propKey + (i === 0 ? '' : '--' + keys[i - 1]);
      return _extends({}, result, (_extends2 = {}, _extends2[prop] = cur, _extends2));
    }, {});
    this.parseObjectProp(obj);
  };
  _proto.parseObjectProp = function parseObjectProp(propValue, meta) {
    if (meta === void 0) {
      meta = {};
    }
    propValue = typeof propValue === 'function' ? propValue(this.props) : propValue;
    var breakpoints = this.config.theme.breakpoints;
    var parsed = parse(propValue, breakpoints);
    for (var _iterator6 = _createForOfIteratorHelperLoose(parsed), _step6; !(_step6 = _iterator6()).done;) {
      var _step6$value = _step6.value,
        selector = _step6$value.selector,
        selectorType = _step6$value.selectorType,
        style = _step6$value.style;
      var entries = Object.entries(style);
      if (!entries.length) continue;
      // entries.length is 1
      var _entries$2 = entries[0],
        propKey = _entries$2[0],
        _propValue = _entries$2[1];
      var option = {
        propKey: propKey,
        propValue: _propValue,
        props: this.props,
        meta: _extends({}, meta)
      };
      // for style props
      if (style && Object.keys(style).length) {
        option.style = style;
      }
      if (selectorType === 'pseudo' && option.meta) {
        var _ref4 = selector.match(/(:+)(.+)/) || [],
          pseudoPrefix = _ref4[1],
          pseudo = _ref4[2];
        option.meta.pseudoPrefix = pseudoPrefix;
        option.meta.pseudo = pseudo;
      }
      if (selectorType === 'child' && option.meta) {
        option.meta.childSelector = selector;
      }
      if (selectorType === 'sibling' && option.meta) {
        option.meta.siblingSelector = selector;
      }
      var atom = new Atom(option, false);
      try {
        this.mutateAtom(atom);
      } catch (error) {
        continue;
      }
      // not match atomic props rule
      if (!atom.style || !Object.keys(atom.style).length) {
        atom.handled = true;
      }
      var cachedAtom = store.atomCache.get(atom.id);
      if (cachedAtom) {
        this.addAtom(cachedAtom);
      } else {
        this.addAtom(atom);
      }
    }
  };
  _proto.makeResponsiveStyle = function makeResponsiveStyle(breakpoint, rule) {
    var breakpoints = this.config.theme.breakpoints;
    return "@media (min-width: " + breakpoints[breakpoint] + ") {" + rule + "}";
  }
  /**
   * get component classNames
   */;
  _proto.getClassNames = function getClassNames() {
    var _this3 = this;
    /**
     * handle override style
     * @example
     * <Box class="red200 blue200"></Box> will get <div class="blue200"></div>
     * <Box class="px2 px4"></Box> will get <div class="px4"></div>
     */
    var classNames = [];
    this.atoms.reduce(function (result, cur) {
      if (!cur.style || !Object.keys(cur.style).length) return result;
      var index = result.findIndex(function (i) {
        return i.styleKeysHash === cur.styleKeysHash;
      });
      if (!cur.isValid) return result;
      var className = _this3.getClassNameById(cur.id);
      if (index === -1) {
        classNames.push(className);
        result = [].concat(result, [cur]);
      } else {
        result.splice(index, 1, cur);
        classNames.splice(index, 1, className);
      }
      return result;
    }, []);
    var _this$props$className = this.props.className,
      className = _this$props$className === void 0 ? '' : _this$props$className;
    var filteredClassNames = className.split(/\s+/).filter(function (i) {
      return !classNames.includes(i) && !!i;
    });
    classNames = classNames.concat(filteredClassNames);
    if (this.hasResponsive) {
      var _classNames;
      var breakpoints = this.config.theme.breakpoints;
      (_classNames = classNames).push.apply(_classNames, Object.keys(breakpoints).map(function (i) {
        return "r-" + i;
      }));
    }
    return classNames;
  }
  /**
   * get style object
   */;
  _proto.toStyle = function toStyle() {
    var _this4 = this;
    var style = this.atoms.reduce(function (result, atom) {
      if (!atom.isValid) return result; // not style type
      var colors = store.theme.colors;
      var style = Object.entries(atom.style).reduce(function (c, _ref5) {
        var _extends3;
        var key = _ref5[0],
          value = _ref5[1];
        var cssValue = _this4.formatCssValue(jsKeyToCssKey(key), colors[value] || value);
        return _extends({}, c, (_extends3 = {}, _extends3[key] = cssValue, _extends3));
      }, {});
      return _extends({}, result, style);
    }, {});
    return style;
  }
  /**
   * get rules for parser.insertRules
   * @returns
   */;
  _proto.toRules = function toRules(enableInserted) {
    if (enableInserted === void 0) {
      enableInserted = false;
    }
    var _this$config$mode$cla = this.config.mode.classPrefix,
      classPrefix = _this$config$mode$cla === void 0 ? '' : _this$config$mode$cla;
    var rules = [];
    var breakpoints = this.config.theme.breakpoints;
    var breakpointKeys = Object.keys(breakpoints);
    for (var _iterator7 = _createForOfIteratorHelperLoose(this.atoms), _step7; !(_step7 = _iterator7()).done;) {
      var atom = _step7.value;
      var rule = '';
      var id = atom.id,
        isValid = atom.isValid,
        _atom$style = atom.style,
        style = _atom$style === void 0 ? {} : _atom$style,
        meta = atom.meta;
      // no style in falsy prop
      if (!isValid) continue;
      // empty style
      if (isEmptyObj(style)) continue;
      var pseudo = meta.pseudo,
        pseudoPrefix = meta.pseudoPrefix,
        mode = meta.mode,
        _meta$breakpoint = meta.breakpoint,
        breakpoint = _meta$breakpoint === void 0 ? '' : _meta$breakpoint,
        childSelector = meta.childSelector,
        siblingSelector = meta.siblingSelector,
        parentClass = meta.parentClass;
      if (!enableInserted) {
        if (atom.inserted) continue;
      }
      atom.inserted = true;
      var className = this.getClassNameById(id);
      var selector = meta.global ? meta.global : "." + className;
      if (mode) selector = "." + classPrefix + mode + " " + selector;
      if (siblingSelector) {
        selector = "" + selector + siblingSelector;
      }
      if (pseudo) {
        var pseudoSelector = pseudoPrefix + pseudo;
        if (parentClass) {
          selector = "." + parentClass + pseudoSelector + " " + selector;
        } else {
          selector = selector + pseudoSelector;
        }
      }
      // TODO: need refactor
      if (childSelector) {
        if (breakpoint) {
          var keys = breakpointKeys.slice(0, breakpointKeys.indexOf(breakpoint) + 1);
          selector = selector + '.' + keys.map(function (i) {
            return "r-" + i;
          }).join('.') + (" " + childSelector);
        } else {
          selector = selector + " " + childSelector;
        }
      } else {
        if (breakpoint) {
          var _keys = breakpointKeys.slice(0, breakpointKeys.indexOf(breakpoint) + 1);
          selector = selector + '.' + _keys.map(function (i) {
            return "r-" + i;
          }).join('.');
        }
      }
      rule = selector + " { " + this.styleToString(style, atom.meta) + " }";
      if (breakpoint) {
        rule = this.makeResponsiveStyle(breakpoint, rule);
      }
      rules.push(rule);
    }
    // console.log('this.atoms:', this.atoms)
    return rules;
  };
  _proto.getParsedProps = function getParsedProps() {
    var _this5 = this;
    var props = this.props,
      atoms = this.atoms;
    if (isEmptyObj(props)) return {};
    var entries = Object.entries(props);
    /** ignore atomic prop */
    var parsedProps = entries.reduce(function (result, _ref6) {
      var key = _ref6[0],
        value = _ref6[1];
      // ignore prop with  postfix
      if (/.*--(\d+)?[a-z]+$/i.test(key)) return result;
      var find = atoms.find(function (atom) {
        return [atom.propKey, atom.key, atom.id].concat(_this5.config.objectPropKeys || []).includes(key) || atom.propKeys.includes(key);
      });
      if (!find) result[key] = value;
      return result;
    }, {});
    return parsedProps;
  };
  _proto.insertRules = function insertRules() {
    var rules = this.toRules();
    styleSheet.insertStyles(rules);
  };
  _createClass(Parser, [{
    key: "uniqueClassName",
    get: function get() {
      return objectToClassName(Object.keys(this.props));
    }
  }, {
    key: "hasResponsive",
    get: function get() {
      return this.atoms.some(function (i) {
        return !!i.meta.breakpoint;
      });
    }
  }, {
    key: "store",
    get: function get() {
      return store;
    }
  }, {
    key: "config",
    get: function get() {
      return store.config;
    }
  }, {
    key: "plugins",
    get: function get() {
      return store.config.plugins;
    }
  }]);
  return Parser;
}();

/**
 * convert atomic props to classNames
 *
 * @param args
 *
 * Valid args:
 *
 * ```js
 * css('toCenter', 'px-8', 'mb2')
 * css('toCenter px-8')
 * css(['toCenter', 'px-8'])
 * css({
 *  px: 8,
 *  mb: 2,
 * })
 * ```
 * @returns  classNames
 */
function css() {
  var _store$config;
  for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
    args[_key] = arguments[_key];
  }
  if (!args.length) return '';
  var parser = new Parser(argsToProps(args, (_store$config = store.config) == null || (_store$config = _store$config.objectPropKeys) == null ? void 0 : _store$config[0]));
  parser.insertRules();
  return parser.getClassNames().join(' ');
}

function isAtomicArgsValid(value) {
  var props = argsToProps([value]);
  var parser = new Parser(props);
  var parsedProps = parser.getParsedProps();
  return !Object.keys(parsedProps).length;
}

function createStyle() {
  for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
    args[_key] = arguments[_key];
  }
  var props = args.reduce(function (result, cur) {
    var _extends2;
    if (typeof cur === 'string') return _extends({}, result, (_extends2 = {}, _extends2[cur] = true, _extends2));
    // not string, is object
    return _extends({}, result, {
      css: cur
    });
  }, {});
  var parser = new Parser(props);
  var style = parser.toStyle();
  return style;
}

/**
 * create keyframe style
 * @param input
 * @returns
 */
function keyframes(input, name) {
  var content = Object.entries(input).reduce(function (result, _ref) {
    var key = _ref[0],
      value = _ref[1];
    var str = Object.entries(value).reduce(function (r, _ref2) {
      var k = _ref2[0],
        v = _ref2[1];
      return r + (jsKeyToCssKey(k) + ": " + v + ";");
    }, '');
    return result + (jsKeyToCssKey(key) + " {" + str + "}");
  }, '');
  var animationName = name || 'keyframes-' + hash$1(JSON.stringify(input));
  var rule = "\n    @keyframes " + animationName + " {\n      " + content + "\n    }\n  ";
  styleSheet.insertStyles([rule]);
  return animationName;
}

/**
 * inject styles to global
 * @param cssObj
 * @exmaple
 * ```js
    injectGlobalStyle({
      '*, ::before, ::after': {
        borderStyle: 'solid',
        borderColor: '#d4d4d4',
        boxSizing: 'border-box',
      },
      body: {
        boxSizing: 'border-box',
      },
    })
 * ```
 */
function injectGlobalStyle(cssObj) {
  var entries = Object.entries(cssObj);
  var parser = new Parser({});
  for (var _i = 0, _entries = entries; _i < _entries.length; _i++) {
    var _entries$_i = _entries[_i],
      key = _entries$_i[0],
      value = _entries$_i[1];
    var atom = new Atom({
      propKey: key,
      propValue: value,
      value: value,
      meta: {
        global: key
      },
      style: value
    });
    parser.addAtom(atom);
  }
  parser.insertRules();
}

// import { injectGlobalStyle } from './injectGlobalStyle'
// injectGlobalStyle({
//   '*, ::before, ::after': {
//     borderWidth: 0,
//     borderStyle: 'solid',
//     borderColor: '#e5e5e5',
//     boxSizing: 'border-box',
//   },
// })
var setConfig = store.setConfig,
  getConfig = store.getConfig,
  setTheme = store.setTheme,
  getTheme = store.getTheme,
  setMode = store.setMode,
  getMode = store.getMode,
  addAtom = store.addAtom,
  composeAtom = store.composeAtom;
var fowerStore = store;

export { Atom, Parser, addAtom, composeAtom, createStyle, css, digitReg, fowerStore, getConfig, getMode, getTheme, injectGlobalStyle, isAtomicArgsValid, keyframes, setConfig, setMode, setTheme, styleSheet };
//# sourceMappingURL=core.esm.js.map
