import { Atom } from './atom';
import { Props } from './typings';
declare type Dict = Record<string, any>;
/**
 * @example p2,mx4,left10,gap4...
 * @example p-20,opacity-80
 */
export declare const digitReg: RegExp;
/**
 * An Abstract tool to handle atomic props
 */
export declare class Parser {
    props: Props;
    constructor(props?: Props);
    /**
     * atom parsed from props
     */
    atoms: Atom[];
    /**
     * For extensibility
     */
    data: Record<string, any>;
    get uniqueClassName(): string;
    get hasResponsive(): boolean;
    get store(): import("./store").Store;
    get config(): import("./typings").Configuration;
    get plugins(): any[];
    getClassNameById: (id: string) => string;
    /**
     * traverse Props to init atoms
     */
    traverseProps(props: Props): void;
    getAutoDarkModeAtom(atom: Atom): Atom | null;
    createAutoDarkModeAtom(): void;
    /**
     * Get final css value
     * @param key css key, eg: font-size, padding-top
     * @param value css value
     * @returns
     */
    formatCssValue(key: string, value: any): any;
    /**
     * convert style object to string
     * @param style
     * @param meta
     * @example
     * { width: 10 } -> "width: 10px;"
     * { paddingTop: 10, paddingBottom: 10 } -> "padding-top: 10px;padding-bottom: 10px;"
     * @returns
     */
    styleToString(style: Dict, meta: Atom['meta']): string;
    addAtom(atom: Atom): void;
    /**
     * prop that can to handle, only primitive value type is valid
     * @param propKey
     * @param propValue
     * @returns
     */
    isValidProp(propKey: string, propValue: any): boolean;
    /**
     * to mutate atom attribute, and add atom to this.atoms
     * @param atom
     */
    mutateAtom(atom: Atom): void;
    parseResponsiveValue(propKey: string, propValue: any[]): void;
    parseObjectProp(propValue: any, meta?: {}): void;
    makeResponsiveStyle(breakpoint: string, rule: string): string;
    /**
     * get component classNames
     */
    getClassNames(): string[];
    /**
     * get style object
     */
    toStyle(): any;
    /**
     * get rules for parser.insertRules
     * @returns
     */
    toRules(enableInserted?: boolean): string[];
    getParsedProps(): any;
    insertRules(): void;
}
export {};
